#!/usr/bin/env python

'''py.test test cases to test OpenStudio.

These tests run based on webclient and need web2py server running.
'''

import datetime
import calendar
from gluon.contrib.populate import populate
from populate_os_tables import populate_customers
    
def prepare_classes(web2py):
    populate_customers(web2py)
    populate(web2py.db.school_locations, 2)
    populate(web2py.db.school_classtypes, 3)
    populate(web2py.db.teachers, 5)
    
    web2py.db.classes.insert(school_locations_id='1', 
                             school_classtypes_id='1', 
                             Week_day='1',
                             Starttime='06:00:00',
                             Endtime='09:00:00',
                             Startdate='2014-01-01',
                             Enddate='2015-01-01',
                             teachers_id='1',
                             teachers_id2='2',
                             Maxstudents=20,
                             DropInPrice='12.50',
                             TrialPrice='7.50'
                             )
    
    web2py.db.classes_registration.insert(customers_id='1', 
                                          classes_id='1')
    web2py.db.classes_waitinglist.insert(customers_id='1',
                                         classes_id='1')
    class_date = web2py.db.classes(1).Startdate
    web2py.db.classes_attendance.insert(customers_id='1', 
                                        classes_id='1', 
                                        ClassDate=class_date, 
                                        AttendanceType='1')
    web2py.db.commit()
    
def populate_school_subscriptions(web2py):
    web2py.db.school_subscriptions.insert(Name="one class a week",
                                     MonthlyFee="40.00", 
                                     NRofClasses='1')
    web2py.db.commit()
    
def test_customers_add(client, web2py):
    '''
        Created a customer?
    '''
    client.get('/customers/edit')
    assert 'name="FirstName"' in client.text
    assert 'name="LastName"' in client.text
    client.get('/customers/edit')
    data = dict(FirstName='Edwin',
                LastName='van de Ven',
                Email='test@openstudioproject.com',
                )
    client.post('/customers/edit', data=data)
    assert client.status == 200

    client.get('/customers')
    assert client.status == 200
    assert data['FirstName'] in client.text
    assert data['LastName'] in client.text
    assert not 'class="error"' in client.text # make sure the form validated
    
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.customers.FirstName == data['FirstName']).count() == 1
    
    
def test_customers_edit(client, web2py):
    '''
        Can we edit a customer
    '''
    populate_customers(web2py)
    assert web2py.db(web2py.db.customers).count() > 0
    
    client.get('/customers')
    assert client.status == 200

    
    first_customer = web2py.db.customers(1)
    fname = first_customer.FirstName
    lname = first_customer.LastName
    
    client.get('/customers/edit/1')
    assert client.status == 200
    assert fname in client.text
    assert lname in client.text

def test_customers_subscription_add(client, web2py):
    '''
        Can we add a customers_subscription?
    '''
    populate_school_subscriptions(web2py)
    populate_customers(web2py)
    assert web2py.db(web2py.db.school_subscriptions).count() > 0
    assert web2py.db(web2py.db.customers).count() > 0
    
    client.get('/customers/subscription_add/1')
    assert client.status == 200
    
    data = dict(school_subscriptions_id='1',
                Startdate='2014-01-01',
                )
    client.post('/customers/subscription_add/1', data=data)
    assert client.status == 200
    assert web2py.db(web2py.db.customers_subscriptions).count() == 1
    
    mstype_name = web2py.db.school_subscriptions(1).Name
    
    client.get('/customers/subscriptions?cuID=1')
    assert client.status == 200
    assert mstype_name in client.text
    
def test_customers_subscription_edit(client, web2py):
    '''
        can we edit a subscription?
    '''
    populate_school_subscriptions(web2py)
    populate_customers(web2py)
    populate(web2py.db.customers_subscriptions, 1)
    
    assert web2py.db(web2py.db.school_subscriptions).count() > 0
    assert web2py.db(web2py.db.customers).count() > 0
    assert web2py.db(web2py.db.customers_subscriptions).count() > 0
    
    client.get('/customers/subscription_edit/1/1')
    assert client.status == 200
    assert web2py.db(web2py.db.customers_subscriptions).count() == 1
    
    data = dict(school_subscriptions_id='1',
                Startdate='2016-01-01',
                )
    client.post('/customers/subscription_edit/1/1', data=data)
    assert client.status == 200
    
    client.get('/customers/subscriptions?cuID=1')
    assert client.status == 200
    assert data['Startdate'] in client.text
    
def test_subscription_pause_resume(client, web2py):
    '''
        Can we pause a subscription?
    '''
    populate_school_subscriptions(web2py)
    populate_customers(web2py, 1)
    populate(web2py.db.customers_subscriptions, 1)
    
    assert web2py.db(web2py.db.school_subscriptions).count() > 0
    assert web2py.db(web2py.db.customers).count() > 0
    assert web2py.db(web2py.db.customers_subscriptions).count() > 0
    
    client.get('/customers/subscriptions?cuID=1')
    assert client.status == 200
    assert 'Pause' in client.text
    
    client.get('/customers/subscription_pause/1')
    assert client.status == 200
    assert 'Current month' in client.text
    
    ## check simple pause and resume
    # pause
    url = '/customers/subscription_pause_current_month/1'
    client.get(url)
    assert client.status == 200
    assert 'Pause subscription' in client.text
    
    data = dict(Description = 'strawberries')
    client.post(url, data=data)
    assert client.status == 200
    assert 'Subscriptions' in client.text # verify redirect
    assert 'Resume' in client.text # verify button text change
    assert web2py.db(web2py.db.customers_subscriptions_paused).count() > 0
    
    client.get('/customers/subscription_pause_history/?cusID=1')
    assert client.status == 200
    assert data['Description'] in client.text
    # resume
    client.get('/customers/subscription_pause/1?resume=1')
    assert client.status == 200
    assert 'Resume' in client.text # check if there's a resume link instead of the pause links
    
    client.get('/customers/subscription_resume/1')
    assert client.status == 200
    assert 'Subscriptions' in client.text
    assert 'Pause' in client.text
    assert web2py.db(web2py.db.customers_subscriptions_paused).count() == 0
    
    ## check multiple month pause and resume
    # pause
    url = '/customers/subscription_pause_multiple_months/1'
    client.get(url)
    assert client.status == 200
    assert 'From' in client.text
    
    data = dict(from_month='1',
                from_year='2014',
                until_month='6',
                until_year='2999',
                description='cherries')
    client.post(url, data)
    assert client.status == 200
    assert 'Subscriptions' in client.text # verify redirection
    assert web2py.db(web2py.db.customers_subscriptions_paused).count() > 0
    
    client.get('/customers/subscription_pause_history/?cusID=1')
    assert client.status == 200
    assert data['description'] in client.text
    # resume 
    ''' When resuming and there's a pause that's started before,
        the end date should be set to the last day of the previous month '''
    client.get('/customers/subscription_resume/2') # increase the counter because of the test above
    assert client.status == 200
    assert 'Subscriptions' in client.text
    assert 'Pause' in client.text
    assert web2py.db(web2py.db.customers_subscriptions_paused).count() == 1
    
    today = datetime.date.today()
    if today.month == 1:
        month = 12
        year = today.year - 1
    else:
        year = today.year
        month = today.month - 1
    last_day_previous_month = datetime.date(year, 
                                            month, 
                                            calendar.monthrange(year,month)[1])
    assert web2py.db(web2py.db.customers_subscriptions_paused.Enddate == \
                     last_day_previous_month).count() == 1
    
def test_classcard_add(client, web2py):
    '''
        can we add a classcard?
    '''
    populate_customers(web2py)
    assert web2py.db(web2py.db.customers).count() > 0
    
    client.get('/customers/classcard_add/1')
    assert client.status == 200
    
    data = dict(Startdate='2014-01-01',
                Enddate='2014-02-01',
                Classes='10',
                Price='125')
    client.post('/customers/classcard_add/1', data=data)
    assert client.status == 200
    
    assert 'Class card' in client.text # check redirection
    assert data['Startdate'] in client.text # check form submission
    assert data['Enddate'] in client.text
    
def test_classcard_edit(client, web2py):
    '''
        can we edit a classcard?
    '''
    populate_customers(web2py)
    populate(web2py.db.customers_classcards, 10)
    web2py.db.commit()
    assert web2py.db(web2py.db.customers).count() > 0
    assert web2py.db(web2py.db.customers_classcards).count() > 0
    
    first_pass = web2py.db().select(web2py.db.customers_classcards.ALL).first()
    url = '/customers/classcard_edit/' + unicode(first_pass.customers_id) + "/" + unicode(first_pass.id)
    
    client.get(url)
    assert client.status == 200
    
    data = dict(Startdate='2016-01-01',
                Enddate='2016-02-01',
                Classes='10',
                Price='125')
    client.post(url, data=data)
    assert client.status == 200
    
    assert 'classcards' in client.text # check redirection
    assert data['Startdate'] in client.text # check form submission
    assert data['Enddate'] in client.text

def test_classes(client, web2py):
    """
        is all the info being shown?
        This tests only if the display works. 
        Actually inputting info will be tested in the classes tests.
    """
    prepare_classes(web2py)
    
    assert web2py.db(web2py.db.customers).count() > 0
    assert web2py.db(web2py.db.classes).count() > 0
    assert web2py.db(web2py.db.classes_registration).count() > 0
    assert web2py.db(web2py.db.classes_waitinglist).count() > 0
    assert web2py.db(web2py.db.classes_attendance).count() > 0
    
    client.get('/customers/classes?cuID=1')
    school_locations_id = web2py.db.classes(1).school_locations_id
    location = web2py.db.school_locations(school_locations_id).Name
    
    check_text = location.split(" ")[0] # it's autogenerated stuff here... so just take the first word
    
    # if the location name occurs 3 times, it's shown under registration, waitinglist and attendance
    assert client.text.count(check_text) == 3
    
def test_classes_add(client, web2py):
    """
        Test if the add a class page actuall shows 
        and then add a class
    """
    prepare_classes(web2py)
    
    client.get('/customers/classes_edit/?cuID=1')
    assert client.status == 200
    assert 'registration' in client.text
    
    """
        now test adding the customer to a class by checking if 
        the location name appears on the /customers/classes page
    """
    
    web2py.db.classes_registration.truncate()
    web2py.db.classes_waitinglist.truncate()
    web2py.db.classes_attendance.truncate()
    web2py.db.customers_subscriptions.truncate()
    web2py.db.commit()
    
    populate_school_subscriptions(web2py)
    # add a subscription
    web2py.db.customers_subscriptions.insert(customers_id='1',
                                             school_subscriptions_id='1', 
                                             Startdate='2014-01-01'
                                             )
    web2py.db.commit()

    school_locations_id = web2py.db.classes(1).school_locations_id
    location = web2py.db.school_locations(school_locations_id).Name
    # it's autogenerated stuff here... so just take the first word
    check_text = location.split(" ")[0] 
    
    client.get('/classes/register?clsID=1&cuID=1')
    assert client.status == 200
    assert 'registration' in client.text # check redirection from classes/register
    assert web2py.db(web2py.db.classes_registration).count() == 1
    
    client.get('/customers/classes?cuID=1')
    assert client.status == 200
    assert check_text in client.text # check if the registration actually happened
    
def test_payments_info_add(client, web2py):
    """
        Can we add info for a customer?
    """
    populate_customers(web2py)
    assert web2py.db(web2py.db.customers).count() > 0
    
    url = '/customers/payment_info_add/1'
    client.get(url)
    assert client.status == 200
    
    data = dict(payments_methods_id='3',
                AccountNumber='123456',
                AccountHolder='Hello',
                BIC="NLBIC123",
                MandateSignatureDate='2014-01-01',
                BankName='ING',
                BankLocation='NL'
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection 
    assert data['AccountNumber'] in client.text
    
    client.get(url) # check if we can add only once
    assert client.status == 200
    assert 'already' in client.text
    
def test_payments_info_edit(client, web2py):
    """
        Can we edit info for a customer?
    """
    populate_customers(web2py, 1)
    
    # get a random url to make sure the setup function runs to populate the 
    # payment_methods table
    
    url = '/customers/index'
    client.get(url)
    assert client.status == 200
    
    # set session.customers_id from edit
    url = '/customers/edit/1'
    client.get(url)
    assert client.status == 200
    
    web2py.db.customers_payment_info.insert(customers_id='1', 
                                            payment_methods_id=3)
    web2py.db.commit()
    assert web2py.db(web2py.db.customers).count() > 0
    assert web2py.db(web2py.db.customers_payment_info).count() == 1
    
    url = '/customers/payment_info_edit/1/1'
    
    client.get(url)
    assert client.status == 200
    
    data = dict(payment_methods_id='3',
                AccountNumber='123456',
                AccountHolder='Hello',
                BIC="NLBIC123",
                MandateSignatureDate='2014-01-01',
                BankName='ING',
                BankLocation='NL'
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection
    assert data['AccountNumber'] in client.text
    
def test_payments_ap_add(client, web2py):
    """
        Can we add an alternative payment?
    """
    today = datetime.date.today()
    populate_customers(web2py, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    
    url = '/customers/alternativepayment_add/1'
    client.get(url)
    assert client.status == 200
    
    assert 'value="' + unicode(today.year) + '"' in client.text # check default year value
    assert 'selected="selected" value="' + unicode(today.month) + '"' in client.text # check default month value
    
    data = dict(PaymentYear=today.year,
                PaymentMonth=today.month,
                Amount='99999',
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection
    assert data['Amount'] in client.text
    
def test_payments_ap_edit(client, web2py):
    """
        Can we edit an alternative payment?
    """
    populate_customers(web2py, 1)
    populate(web2py.db.payment_categories, 4)
    populate(web2py.db.alternativepayments, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.alternativepayments).count() > 0
    
    url = '/customers/alternativepayment_edit/1/1'
    client.get(url)
    assert client.status == 200
    
    data = dict(PaymentYear='2015',
                PaymentMonth='1',
                Amount='99999',
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection
    assert data['Amount'] in client.text
    
def test_payments_op_add(client, web2py):
    '''
        Can we add an overdue payment?
    '''
    today = unicode(datetime.date.today())
    populate_customers(web2py, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    
    url = '/customers/overduepayment_add/1'
    client.get(url)
    assert client.status == 200
    assert today in client.text # check default value for date
    
    data = dict(Paid='on',
                PaymentDate=today,
                Amount='999912345',
                Description="Apple pie with almonds is sooo nice!"
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection 
    assert data['Amount'] in client.text
    
def test_payments_op_edit(client, web2py):
    '''
        Can we edit an overdue payment?
    '''
    today = unicode(datetime.date.today())
    populate_customers(web2py, 1)
    web2py.db.overduepayments.insert(customers_id='1', Paid=False, PaymentDate=today, Amount='20', Description="Cherry pie is also very nice!")
    web2py.db.commit()
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.overduepayments).count() == 1
    
    url = '/customers/overduepayment_edit/1/1'
    client.get(url)
    assert client.status == 200
    assert "Cherry" in client.text
    
    data = dict(Paid='on',
                PaymentDate=today,
                Amount='999912345',
                Description="Apple pie with almonds is sooo nice!"
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection 
    assert data['Amount'] in client.text

def test_payment_overview_add(client, web2py):
    '''
        Can we add a payment overview?
    '''
    today = datetime.date.today()
    populate_customers(web2py, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    
    url = '/customers/paymentoverview_add/1'
    client.get(url)
    assert client.status == 200
    
    data = dict(PaymentYear=today.year,
                M01='on'
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection 
    assert 'checked="checked"' in client.text
    
def test_payment_overview_edit(client, web2py):
    '''
        Can we edit a payment overview?
    '''
    today = datetime.date.today()
    populate_customers(web2py, 1)
    web2py.db.paymentsummary.insert(customers_id='1', 
                                    PaymentYear=today.year,
                                    M01=False,
                                    M02=False,
                                    M03=False,
                                    M04=False,
                                    M05=False,
                                    M06=False,
                                    M07=False,
                                    M08=False,
                                    M09=False,
                                    M10=False,
                                    M11=False,
                                    M12=False
                                    )
    web2py.db.commit()
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.paymentsummary).count() == 1
    
    url = '/customers/paymentoverview_edit/1/1'
    client.get(url)
    assert client.status == 200
    
    data = dict(PaymentYear=today.year,
                M01='on'
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "info" in client.text # check redirection 
    assert 'checked="checked"' in client.text

def test_memo_add(client, web2py):
    '''
        Created a memo?
    '''
    populate_customers(web2py, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    
    # first get this url to set the session variable for redirection
    client.get('/customers/memos?cuID=1') 
    assert client.status == 200
    
    url = '/customers/memo_add/1'
    client.get(url)
    assert client.status == 200
    data = dict(Title='Bananas',
                Msg='message text',
                Startdate='2014-01-01',
                Priority='2',
                Visible='on',
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "Memos" in client.text # check is we're at the customers page
    
    assert web2py.db(web2py.db.memos).count() == 1
    assert data['Title'] in client.text
    
    assert web2py.db(web2py.db.memos.Title == data['Title']).count() == 1
    assert web2py.db(web2py.db.memos.customers_id == 1).select(web2py.db.memos.ALL) != None
    
def test_memo_edit(client, web2py):
    '''
        Can we edit a memo
    '''
    populate_customers(web2py, 1)
    web2py.db.memos.insert(customers_id=1, Visible=True, Title="Strawberries", Msg="I Like them!!", Startdate='2014-01-01', Priority=2)
    web2py.db.commit()
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.memos).count() == 1
    
    
    client.get('/customers/memos?cuID=1') # first get this url to set the session variable for redirection
    assert client.status == 200
    
    url = '/customers/memo_edit/1/1'
    client.get(url)
    assert client.status == 200
    data = dict(Title='Bananas',
                Msg='message text',
                Startdate='2014-01-01',
                Priority='2',
                Visible='on',
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "Memos" in client.text # check if we're at the customers page
    
    assert data['Title'] in client.text
    
    assert web2py.db(web2py.db.memos.Title == data['Title']).count() == 1
    assert web2py.db(web2py.db.memos.customers_id == 1).select(web2py.db.memos.ALL) != None
    
def test_documents(client, web2py):
    '''
        Can we get a list of the documents available?
    '''
    populate_customers(web2py, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    
    url = '/customers/documents?cuID=1'
    client.get(url)
    assert client.status == 200
    assert "Documents" in client.text
