#!/usr/bin/env python

'''py.test test cases to test OpenStudio.

These tests run based on webclient and need web2py server running.
'''

import datetime
import gluon.contrib.simplejson as sj
from gluon.contrib.populate import populate
from populate_os_tables import populate_classes
from populate_os_tables import populate_customers

def test_class_add(client, web2py):
    '''
        Can we add a class?
    '''
    populate(web2py.db.school_locations, 1)
    populate(web2py.db.school_classtypes, 1)
    populate(web2py.db.teachers, 1)
    
    url = '/classes/class_add'
    client.get(url)
    assert client.status == 200
    data = dict(school_locations_id='1',
                school_classtypes_id='1',
                Week_day='1',
                Starttime='06:00:00',
                Endtime='08:00:00',
                Startdate='2014-01-01',
                teachers_id='1',
                Maxstudents='123456',
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "Schedule" in client.text # check redirection to schedule
    
    assert web2py.db(web2py.db.classes).count() == 1
    assert data['Maxstudents'] in client.text   
    assert web2py.db(web2py.db.classes.Maxstudents == data['Maxstudents']).count() == 1
    
def test_class_edit(client, web2py):
    '''
        Can we edit a class?
    '''
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    url = '/classes/class_edit/1'
    client.get(url)
    assert client.status == 200
    data = dict(school_locations_id='1',
                school_classtypes_id='1',
                Week_day='1', # Monday
                Starttime='06:00:00',
                Endtime='08:00:00',
                Startdate='2014-01-01',
                teachers_id='1',
                Maxstudents='99999',
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "Schedule" in client.text # check redirection to schedule
    
    assert data['Maxstudents'] in client.text   
    assert web2py.db(web2py.db.classes.Maxstudents == data['Maxstudents']).count() == 1
    
def test_schedule(client, web2py):
    '''
        Is the schedule showing all things as it should?
    '''
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    url = '/classes/schedule'
    client.get(url)
    assert client.status == 200
    assert 'Schedule' in client.text
    
    url = '/classes/schedule?year=2014&week=2' # we need the second week because the class is on a monday and the in week 1 monday is in december 2013.
    client.get(url)
    assert client.status == 200
    assert 'Manage' in client.text
    assert 'Edit' in client.text
    assert 'Remove' in client.text
    location_check = web2py.db.school_locations(1).Name.split(' ')[1]
    assert location_check in client.text
    
def test_status_set_subteacher(client, web2py):
    '''
        Can we add substitute teachers for a class?
    '''
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    classes_id = '1'
    classdate = '2014-01-06'
    
    url  = '/classes/manage/' + classes_id + '/' + classdate
    client.get(url)
    
    url = '/classes/status_set_subteacher/' + classes_id + '/' + classdate
    client.get(url)
    assert client.status == 200
    teachers_id = '1'
    data = dict(teachers_id=teachers_id, 
                teachers_id2='')
    client.post(url, data=data)
    assert client.status == 200
    # check redirection to the class details page
    assert "Manage" in client.text 
    # check listing of teachers on manage page
    assert 'Substitute teachers' in client.text 
    teacher_name_check = web2py.db.teachers(teachers_id).FirstName.split(" ")[1]
    assert teacher_name_check in client.text
    
    assert web2py.db(web2py.db.classes_subteachers).count() == 1
    
    # again visit the page and check wether the teacher is selected in the select
    url = '/classes/status_set_subteacher/' + classes_id + '/' + classdate
    client.get(url)
    assert client.status == 200
    assert '<option selected="selected" value="1">' in client.text
    
def test_registration(client, web2py):
    '''
        Does the expected attendance page show?
    '''
    populate_classes(web2py)
    populate_customers(web2py, 1)
    web2py.db.customers.insert(FirstName='Edwin',
                               LastName='van de Ven',
                               Gender='Male',
                               Email='edwin@openstudioproject.com'
                               )
    web2py.db.classes_registration.insert(classes_id='1', customers_id='1')
    web2py.db.commit()
    
    assert web2py.db(web2py.db.customers).count() == 2
    assert web2py.db(web2py.db.classes).count() == 1
    assert web2py.db(web2py.db.classes_registration).count() == 1
    
    url = '/classes/attendance/1/2014-01-06'
    client.get(url)
    assert client.status == 200
    assert 'Edwin' in client.text
    
def test_attendance_remove(client, web2py):
    '''
        Can we remove attendance using JSON after adding it?
    '''
    populate_classes(web2py)
    classdate = '2014-01-06'
    web2py.db.customers.insert(FirstName='Edwin',
                               LastName='van de Ven',
                               Gender='Male',
                               Email='edwin@openstudioproject.com')
    web2py.db.classes_attendance.insert(customers_id=1,
                                        classes_id=1,
                                        ClassDate=classdate,
                                        AttendanceType=None,
                                        customers_classcards_id=None)
    web2py.db.commit()
    assert web2py.db(web2py.db.classes_attendance).count() > 0
    
    # first get the manage page to set some session variables
    url = '/classes/manage/1/2014-01-06'
    client.get(url)
    assert client.status == 200
    
    # now try to remove a customer from the attendance list
    url = '/classes/attendance_remove.json'
    
    data = dict(cuID=1,
                clsID=1,
                date=classdate)
    client.post(url, data=data)
    assert client.status == 200
    
    # check whether everything was ok
    json = sj.loads(client.text)
    assert json['status'] == 'success'
    
    
def test_classes_override_attendance(client, web2py):
    '''
        Does the override attendance page work?
    '''
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    url = '/classes/attendance_override/1/2014-01-06'
    client.get(url)
    assert client.status == 200
    
    data = dict(Amount='123456')
    client.post(url, data=data)
    assert client.status == 200
    assert 'Manage' in client.text # check redirection to manage page
    query = (web2py.db.classes_attendance_override.Amount == \
             data['Amount'])
    assert web2py.db(query).count() == 1 
    
    # now can we edit it
    
    client.get(url)
    assert client.status == 200
    assert data['Amount'] in client.text
    
    update_data = dict(Amount='987654')
    client.post(url, data=update_data)
    assert client.status == 200
    assert 'Manage' in client.text # check redirection to manage page
    query = (web2py.db.classes_attendance_override.Amount == \
             update_data['Amount'])
    assert web2py.db(query).count() == 1 
    
def test_status_set_cancelled(client, web2py):
    '''
        Does the cancel status work?
    '''
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    classes_id = '1'
    classdate = '2014-01-06'
    
    url = '/classes/manage/' + classes_id + '/' + classdate
    client.get(url)
    assert client.status == 200
    
    # check cancelling class
    cancel_url = '/classes/status_set_cancelled/1/' + classdate
    client.get(cancel_url)
    assert client.status == 200
    assert 'Manage' in client.text # verify redirection to class details page
    
    assert '<li class="web2py-menu-active">Cancelled</li>' in client.text
    
    query = (web2py.db.classes_cancelled.classes_id == classes_id) & \
            (web2py.db.classes_cancelled.ClassDate == classdate)
    
    assert web2py.db(query).count() == 1
    
    
def test_status_set_open(client, web2py):
    '''
        Does the open status work?
    '''
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    classes_id = '1'
    classdate = '2014-01-06'
    
    # check setting the status as open
    cancel_url = '/classes/status_set_open/' + classes_id + '/' + classdate
    client.get(cancel_url)
    assert client.status == 200
    assert 'Manage' in client.text # verify redirection to class details page
    
    query = (web2py.db.classes_open.classes_id == classes_id) & \
            (web2py.db.classes_open.ClassDate == classdate)
    
    assert web2py.db(query).count() == 1
    
    assert '<li class="web2py-menu-active">Open</li>' in client.text
    
    
def test_teacher_holidays(client, web2py):
    '''
        Check whether all classes in a specified period of time are put in the
        classes_open table, changing their status to open.
    '''
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    url = '/classes/teacher_holidays'
    client.get(url)
    assert client.status == 200
    
    ''' There are 4 Mondays in January 2014, so 4 entries should appear in the 
        classes_open table. '''
    data = dict(teachers_id=1,
                 Startdate='2014-01-01',
                 Enddate='2014-02-01') 
    client.post(url, data=data)
    assert client.status == 200
    # verify redirection (Can't use 'Schedule' because it's int the menu)
    assert 'Monday' in client.text
    # verify that there are 4 entries in the classes_open table
    assert web2py.db(web2py.db.classes_open.id > 0).count() == 4
    
def test_classes_open(client, web2py):
    '''
        Check if classes appear on the classes_open list
    '''
    def next_weekday(d, weekday):
        days_ahead = weekday - d.weekday()
        if days_ahead <= 0: # Target day already happened this week
            days_ahead += 7
        return d + datetime.timedelta(days_ahead)
    
    populate_classes(web2py)
    assert web2py.db(web2py.db.classes).count() == 1
    
    # get 2 mondays later than today
    delta = datetime.timedelta(days=7)
    today = datetime.date.today()
    next_monday = next_weekday(today, 0) # 0 = Monday, 1=Tuesday, 2=Wednesday...
    monday_after_that = next_monday + delta
    
    web2py.db.classes_open.insert(classes_id=1,
                                  ClassDate=next_monday)
    web2py.db.classes_open.insert(classes_id=1,
                                  ClassDate=monday_after_that)
    web2py.db.classes_open.insert(classes_id=1,
                                  ClassDate='2014-01-06')
    web2py.db.commit()
    assert web2py.db(web2py.db.classes_open).count() == 3
    
    # get the page
    url = '/classes/classes_open'
    client.get(url)
    assert client.status == 200
    
    # past open classes shouldn't be shown
    assert '2014-01-06' not in client.text 
    # future ones should be shown
    assert unicode(next_monday) in client.text
    assert unicode(monday_after_that) in client.text
    
