#!/usr/bin/env python

'''py.test test cases to test OpenStudio.

These tests run based on webclient and need web2py server running.
'''

from gluon.contrib.populate import populate
from populate_os_tables import populate_customers

def populate_ws_attendance(web2py):
    '''
        Add a customer, workshop, workshop activity and add the customer to the
        activity
    '''
    populate_customers(web2py, 1)
    populate(web2py.db.workshops, 1)
    populate(web2py.db.workshops_activities, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.workshops).count() == 1
    assert web2py.db(web2py.db.workshops_activities).count() == 1
    
    web2py.db.workshops_registration.insert(workshops_id=1,
                                            customers_id=1)
    web2py.db.commit()
    assert web2py.db(web2py.db.workshops_registration).count() == 1

def test_workshop_index(client, web2py):
    '''
        Is the index page showing?
    '''
    url = '/workshops/index'
    client.get(url)
    assert client.status == 200
    assert 'Workshops' in client.text

def test_workshop_add(client, web2py):
    '''
        Can we add a workshop?
    '''
    url = '/workshops/workshop_add'
    client.get(url)
    assert client.status == 200
    
    data = dict(Name="Mysore_week",
                Startdate='2014-01-01',
                Teacher='Edwin van de Ven',
                Price='250'
                )
    client.post(url, data=data)
    assert client.status == 200
    assert 'Workshops' in client.text # verify redirection
    assert data['Name'] in client.text
    
    assert web2py.db(web2py.db.workshops).count() == 1

def test_workshop_edit(client, web2py):
    '''
        Can we edit a workshop?
    '''
    populate(web2py.db.workshops, 1)
    assert web2py.db(web2py.db.workshops).count() == 1
    
    url = '/workshops/workshop_edit/1'
    client.get(url)
    assert client.status == 200
    
    data = dict(Name="Mysore_week",
                Startdate='2014-01-01',
                Teacher='Edwin van de Ven',
                Price='250'
                )
    client.post(url, data=data)
    assert client.status == 200
    assert 'Workshops' in client.text # verify redirection
    assert data['Name'] in client.text
    
    assert web2py.db(web2py.db.workshops).count() > 0
    
def test_ws_activity_add(client, web2py):
    '''
        Can we add a activity to a workshop?
    '''
    populate(web2py.db.workshops, 1)
    assert web2py.db(web2py.db.workshops).count() == 1

    client.get('/workshops/attendance?wsID=1') # fetch overview page fist to set session var
    assert client.status == 200
    
    url = '/workshops/activity_add/1'
    client.get(url)
    assert client.status == 200
    assert 'Add activity' in client.text
    
    data = dict(Activity="Cherries",
                Activitydate='2014-01-01',
                Starttime='09:00:00',
                Endtime='11:00:00',
                Spaces='20',
                Price='250')
    client.post(url, data=data)
    assert client.status == 200
    assert 'Workshop' in client.text # verify redirection back to workshop main page
    assert data['Activity'] in client.text # is the activity showing in the list?
    
    assert web2py.db(web2py.db.workshops_activities).count() == 1
    
def test_ws_activity_edit(client, web2py):
    '''
        Can we edit an activity?
    '''
    populate(web2py.db.workshops, 1)
    populate(web2py.db.workshops_activities, 1)
    assert web2py.db(web2py.db.workshops).count() == 1
    assert web2py.db(web2py.db.workshops_activities).count() == 1
    
    # fetch overview page fist to set session var
    client.get('/workshops/attendance?wsID=1') 
    assert client.status == 200

    url = '/workshops/activity_edit/1'
    client.get(url)
    assert client.status == 200
    
    data = dict(id='1',
                Activity="Cherries",
                Activitydate='2014-01-01',
                Starttime='09:00:00',
                Endtime='11:00:00',
                Spaces='20',
                Price='250')
    client.post(url, data=data)
    assert client.status == 200
    # verify redirection back to workshop main page
    assert 'Activities' in client.text 
    # is the activity showing in the list?
    assert data['Activity'] in client.text 
    # check the db
    assert web2py.db(web2py.db.workshops_activities).count() > 0
    
def test_ws_activity_duplicate(client, web2py):
    '''
        Is the duplication working?
    '''
    populate(web2py.db.workshops, 1)
    populate(web2py.db.workshops_activities, 1)
    assert web2py.db(web2py.db.workshops).count() == 1
    assert web2py.db(web2py.db.workshops_activities).count() == 1
    
    url = '/workshops/activity_duplicate/1'
    client.get(url)
    assert client.status == 200
    assert 'duplicated' in client.text # is the flash message showing
    assert web2py.db(web2py.db.workshops_activities).count() == 2
    assert web2py.db.workshops_activities(1).Activity == web2py.db.workshops_activities(2).Activity
    
def test_ws_attendance_add(client, web2py):
    '''
        Can we add attendance for the full workshop
    '''
    populate_customers(web2py, 1)
    populate(web2py.db.workshops, 1)
    populate(web2py.db.workshops_activities, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.workshops).count() == 1
    assert web2py.db(web2py.db.workshops_activities).count() == 1
    
    # first go to the workshops page to set some session vars we need later
    client.get('/workshops/attendance?wsID=1')
    # then go to the attendance_add page
    client.get('/workshops/attendance_add')
    # now try to add a customer to the attendance list
    url = '/workshops/change_attendance?customers_id=1'
    client.get(url)
    assert client.status == 200
    assert 'attendance' in client.text.lower()
    
    assert web2py.db(web2py.db.workshops_registration).count() == 1
    
def test_ws_attendance_delete(client, web2py):
    '''
        Can we remove a customer from the attendance list
    '''
    populate_ws_attendance(web2py)
    
    url = '/workshops/attendance_delete.json'
    
    data = dict(id='1')
    client.post(url, data=data)
    assert client.status == 200
    print client.text
    
    
    assert web2py.db(web2py.db.workshops_registration).count() == 0
    
def test_wsa_attendance_add(client, web2py):
    '''
        Can we add attendance for a workshop activity?
    '''
    populate_customers(web2py, 1)
    populate(web2py.db.workshops, 1)
    populate(web2py.db.workshops_activities, 1)
    assert web2py.db(web2py.db.customers).count() == 1
    assert web2py.db(web2py.db.workshops).count() == 1
    assert web2py.db(web2py.db.workshops_activities).count() == 1
    
    # first go to the workshops page to set some session vars we need later
    client.get('/workshops/attendance?wsID=1')
    assert client.status == 200
    # then go to the attendance_add page
    client.get('/workshops/activity_attendance?wsaID=1')
    assert client.status == 200
    client.get('/workshops/activity_attendance_add')
    assert client.status == 200
    # now try to add a customer to the attendance list
    url = '/workshops/change_activity_attendance?customers_id=1&wsaID=1'
    client.get(url)
    assert client.status == 200
    assert 'activity' in client.text.lower()
    
    assert web2py.db(web2py.db.workshops_registration).count() == 1
    
def test_ws_attendance_update_info(client, web2py):
    '''
        Test as JSON
        Is the attendance info ajaj backend working?
    '''
    populate_ws_attendance(web2py)
    
    url = '/workshops/attendance_update_info.json'
    data = dict(id='1',
                Info='on',
                PaymentConfirmation='on')
    client.post(url, data=data)
    assert client.status == 200
    assert 'success' in client.text
    
    assert web2py.db.workshops_registration(1).Info == True
    assert web2py.db.workshops_registration(1).PaymentConfirmation == True
    
    # unset info and payment confirmation. 
    url = '/workshops/attendance_update_info.json'
    data = dict(id='1')
    client.post(url, data=data)
    assert client.status == 200
    assert 'success' in client.text
    
    assert web2py.db.workshops_registration(1).Info == False
    assert web2py.db.workshops_registration(1).PaymentConfirmation == False
    
    
    
def test_ws_memos(client, web2py):
    '''
        Is the memos page shown under workshops?
    '''
    populate(web2py.db.workshops, 1)
    assert web2py.db(web2py.db.workshops).count() == 1
    
    # first get the attendance page to set the session variable 
    # session.wsID
    
    client.get('/workshops/attendance?wsID=1')
    assert client.status == 200
    
    
    url = '/workshops/memos'
    client.get(url)
    assert client.status == 200
    assert 'memos' in client.text.lower()
    
def test_ws_memo_add(client, web2py):
    '''
        Can we add a memo from workshops?
    '''
    populate(web2py.db.workshops, 1)
    assert web2py.db(web2py.db.workshops).count() == 1
    
    # first get the attendance page to set the session variable 
    # session.wsID
    
    client.get('/workshops/attendance?wsID=1')
    assert client.status == 200
    
    url = '/workshops/memo_add'
    client.get(url)
    assert client.status == 200
    
    data = dict(Title='Bananas',
            Msg='message text',
            Startdate='2014-01-01',
            Priority='2',
            Visible='on',
            )
    client.post(url, data=data)
    assert client.status == 200
    assert "Memos" in client.text # check if we're at the workshops memo page
    
    assert web2py.db(web2py.db.memos).count() == 1
    assert data['Title'] in client.text
    
    assert web2py.db(web2py.db.memos.Title == data['Title']).count() == 1
    query = (web2py.db.memos.workshops_id == 1)
    assert not web2py.db(query).select(web2py.db.memos.ALL) is None
    
    # check if it's on the pin board
    url = '/default/pinboard'
    client.get(url)
    assert client.status == 200
    assert data['Title'] in client.text
    
    
def test_ws_memo_edit(client, web2py):
    '''
        Can we edit a memo from workshops?
    '''
    populate(web2py.db.workshops, 1)
    web2py.db.memos.insert(workshops_id=1,
                           Visible=True, 
                           Title="Strawberries", 
                           Msg="I Like them!!", 
                           Startdate='2014-01-01', 
                           Priority=2)
    web2py.db.commit()
    assert web2py.db(web2py.db.workshops).count() == 1
    assert web2py.db(web2py.db.memos).count() == 1
    
    # first get the attendance page to set the session variable 
    # session.wsID
    
    client.get('/workshops/attendance?wsID=1')
    assert client.status == 200
    
    url = '/workshops/memo_edit/1'
    client.get(url)
    assert client.status == 200
    data = dict(Title='Bananas',
                Msg='message text',
                Startdate='2014-01-01',
                Priority='2',
                Visible='on',
                )
    client.post(url, data=data)
    assert client.status == 200
    assert "Memos" in client.text # check if we're at the workshop memos page
    
    assert data['Title'] in client.text
    
    assert web2py.db(web2py.db.memos.Title == data['Title']).count() == 1
    query = (web2py.db.memos.workshops_id == 1)
    assert not web2py.db(query).select(web2py.db.memos.ALL) is None
