#!/usr/bin/env python

'''py.test test cases to test OpenStudio.

These tests run based on webclient and need web2py server running.
'''

from gluon.contrib.populate import populate

def test_users_index(client, web2py):
    '''
        Is the index page showing?
    '''
    url = '/settings/users'
    client.get(url)
    assert client.status == 200
    assert 'OpenStudio users' in client.text
    
def test_user_add(client, web2py):
    '''
        Can we add a user?
    '''
    url = '/settings/user_add'
    client.get(url)
    assert client.status == 200
    assert 'New user' in client.text
    
    data = dict(first_name='Edwin',
                 last_name='van de Ven',
                 email='edwin@openstudioproject.com',
                 password='super_secret')
    client.post(url, data=data)
    assert client.status == 200
    assert 'OpenStudio users' in client.text # verify redirection
    assert data['first_name'] in client.text
    
    assert web2py.db(web2py.db.auth_user).count() == 2
    
    # check if group gets created and user becomes a member of this group
    user_id = 2
    group_id = web2py.auth.user_group(user_id)
    assert not group_id is None # see if a group was created
    assert web2py.auth.has_membership(group_id, user_id) # see if the created user is a member of the created group
    
def test_user_edit(client, web2py):
    '''
        Can we edit a user?
    '''
    populate(web2py.db.auth_user, 1)
    assert web2py.db(web2py.db.auth_user).count() == 1
    
    url = '/settings/user_edit/1'
    client.get(url)
    assert client.status == 200
    assert 'Edit user' in client.text
    
    data = dict(first_name='Edwin',
                 last_name='van de Ven',
                 email='edwin@openstudioproject.com',
                 password='super_secret')
    client.post(url, data=data)
    assert client.status == 200
    assert 'OpenStudio users' in client.text # verify redirection
    assert data['first_name'] in client.text
    
    assert web2py.db(web2py.db.auth_user).count() > 1

def test_payment_categories_index(client, web2py):
    '''
        Is the index page showing?
    '''
    url = '/settings/payment_categories'
    client.get(url)
    assert client.status == 200
    assert 'Payment categories' in client.text
    
def test_payment_categories_add(client, web2py):
    '''
        Can we add a payment category?
    '''
    url = '/settings/paymentcategory_add'
    client.get(url)
    assert client.status == 200
    assert 'New payment category' in client.text
    
    data = dict(Name='Invoices',
                 CategoryType='1')
    client.post(url, data=data)
    assert client.status == 200
    assert 'Payment categories' in client.text # verify redirection
    assert data['Name'] in client.text
    
    assert web2py.db(web2py.db.payment_categories).count() == 1
    
def test_payment_categories_edit(client, web2py):
    '''
        Can we edit a payment category?
    '''
    populate(web2py.db.payment_categories, 1)
    assert web2py.db(web2py.db.payment_categories).count() == 1
    
    url = '/settings/paymentcategory_edit/1'
    client.get(url)
    assert client.status == 200
    assert 'Edit payment category' in client.text
    
    data = dict(Name='Invoices',
                 CategoryType='1')
    client.post(url, data=data)
    assert client.status == 200
    assert 'Payment categories' in client.text # verify redirection
    assert data['Name'] in client.text
    
    assert web2py.db(web2py.db.payment_categories).count() > 0
    
def test_api_users_index(client, web2py):
    '''
        Can we list the API users?
    '''
    web2py.db.sys_api_users.insert(ActiveUser=True,
                                   Username='magicdesign',
                                   APIKey='1234567890',
                                   Description='magic design user')
    web2py.db.commit()
    
    url = '/settings/api_users'
    client.get(url)
    assert client.status == 200
    assert 'API users' in client.text
    assert 'magicdesign' in client.text
    
def test_api_user(client, web2py):
    '''
        Can we add an api user?
    '''
    url = '/settings/api_user_add'
    client.get(url)
    assert client.status == 200
    assert 'New API user' in client.text
    
    data = dict(Username='magicdesign',
                 Description='Magic design user')
    client.post(url, data=data)
    assert client.status == 200
    assert 'API users' in client.text # verify redirection
    assert data['Username'] in client.text
    
    assert web2py.db(web2py.db.sys_api_users).count() == 1
    
    # check whether a key was generated
    api_key = web2py.db.sys_api_users(1).APIKey
    assert api_key != '' and not api_key is None
    
    ## check whether a new key can be generated
    url = '/settings/api_user_generate_new_key/1'
    client.get(url)
    assert client.status == 200
    # check redirection
    assert 'API users' in client.text
    # check whether a new key was generated
    api_new_key = web2py.db.sys_api_users(1).APIKey
    assert api_key != api_new_key
    
    ## check whether deactivating and activating again works
    # check current state (should be active by default)
    assert web2py.db.sys_api_users(1).ActiveUser == True
    # change the state
    url='/settings/api_user_change_active_state/1'
    client.get(url)
    assert client.status == 200
    assert 'API users' in client.text
    assert web2py.db.sys_api_users(1).ActiveUser == False
    # change the state back
    url='/settings/api_user_change_active_state/1'
    client.get(url)
    assert client.status == 200
    assert 'API users' in client.text
    assert web2py.db.sys_api_users(1).ActiveUser == True
    
