#!/usr/bin/env python

'''
    py.test test cases to test the API controller (api.py)
'''

import urllib
import gluon.contrib.simplejson as sj
from gluon.contrib.populate import populate

base_url = 'http://dev.openstudioproject.com:8000'

def populate_schedule(web2py):
    '''
        Sets up a class and API user to test with
    '''
    web2py.db.teachers.insert(FirstName='Edwin',
                              LastName='van de Ven',
                              Gender='Male',
                              Email='edwin@openstudioproject.com')
    web2py.db.teachers.insert(FirstName='Pietje',
                              LastName='Puk',
                              Gender='Male',
                              Email='pietje@puk.nl')
    web2py.db.teachers.insert(FirstName='Aimee',
                              LastName='Garcias',
                              Gender='Female',
                              Email='aimee@ashtanga.sp')
    web2py.db.school_classtypes.insert(Name='Mysore')
    web2py.db.school_classtypes.insert(Name='Led class')
    web2py.db.school_locations.insert(Name="Sittard",
                                      AllowAPI="True")
    web2py.db.school_levels.insert(Name='Level 1')
    web2py.db.classes.insert(school_locations_id=1,
                             school_classtypes_id=1,
                             school_levels_id=1,
                             Week_day=1,
                             Starttime='06:00:00',
                             Endtime='09:00:00',
                             Startdate='2014-01-01',
                             Enddate='',
                             teachers_id=1,
                             teachers_id2=2,
                             Maxstudents=20,
                             DropInPrice=15,
                             TrialPrice=10,
                             AllowAPI=True)
    web2py.db.classes_subteachers.insert(classes_id=1,
                                         ClassDate='2014-01-13',
                                         teachers_id=3,
                                         teachers_id2=2)
    web2py.db.classes_cancelled.insert(classes_id=1,
                                       ClassDate='2014-01-20')
    web2py.db.school_holidays.insert(Description='Carnavals vakantie',
                                     Startdate='2014-01-27',
                                     Enddate='2014-01-30',
                                     Classes=True)
    web2py.db.sys_api_users.insert(ActiveUser=True,
                                   Username='test',
                                   APIKey='test',
                                   Description='test user')
    web2py.db.commit()
    

def test_schedule_get_extension_error():
    '''
        Check wether we get a value error when a call is made without variables
    '''
    url = base_url + '/api/schedule_get'
    page = urllib.urlopen(url).read()
    assert "Extension error" in page
    
def test_schedule_get_authentication_error():
    '''
        Check wether we get an authentication error when we don't supply
        a username and password or a wrong username and password
    '''
    url = base_url + \
        '/api/schedule_get.json?user=test&key=test&year=2014&week=1'
    page = urllib.urlopen(url).read()
    assert "Authentication error" in page
    
def test_schedule_get_value_error():
    '''
        Check wether we specify a string for year and week
    '''
    url = base_url + \
        '/api/schedule_get.json?user=test&key=test&year=bla&week=bla'
    page = urllib.urlopen(url).read()
    assert "Value error" in page
    
def test_schedule_get_missing_value():
    '''
        Check wether we specify a string for year and week
    '''
    url = base_url + \
        '/api/schedule_get.json'
    page = urllib.urlopen(url).read()
    assert "Missing value" in page
    
def test_schedule_get_json(client, web2py):
    '''
        Check wether we can get the information in the database through the api
        using the JSON interface.
    '''
    populate_schedule(web2py)
    
    url = base_url + \
        '/api/schedule_get.json?user=test&key=test&year=2014&week=2'
    page = urllib.urlopen(url).read()
    
    json = sj.loads(page)
    
    # check locations info
    assert json['data']['locations'][0]['id'] == 1
    assert json['data']['locations'][0]['Name']== 'Sittard'
    
    # check classtypes info
    assert json['data']['classtypes'][0]['Name'] == 'Mysore'
    assert json['data']['classtypes'][1]['Name']== 'Led class'
    
    # check teachers info
    assert json['data']['teachers'][0]['name'] == 'Edwin van de Ven'
    assert json['data']['teachers'][1]['name'] == 'Pietje Puk'
    
    # check classes info
    assert json['data']['classes']['Monday']['date'] == '2014-01-06'
    assert not json['data']['classes']['Monday']['holiday']
    assert json['data']['classes']['Monday']['classes'][0]['LocationID'] == 1
    assert json['data']['classes']['Monday']['classes'][0]['Location'] == \
        'Sittard'
    assert json['data']['classes']['Monday']['classes'][0]['ClassTypeID'] == 1
    assert json['data']['classes']['Monday']['classes'][0]['ClassType'] == \
        'Mysore'
    assert json['data']['classes']['Monday']['classes'][0]['Starttime'] == \
        '06:00:00'
    assert json['data']['classes']['Monday']['classes'][0]['Endtime'] == \
        '09:00:00'
    assert json['data']['classes']['Monday']['classes'][0]['TeacherID'] == 1
    assert json['data']['classes']['Monday']['classes'][0]['Teacher'] == \
        'Edwin van de Ven'
    assert json['data']['classes']['Monday']['classes'][0]['TeacherID2'] == 2
    assert json['data']['classes']['Monday']['classes'][0]['LevelID'] == 1
    assert json['data']['classes']['Monday']['classes'][0]['Level'] == \
        'Level 1'
    assert not json['data']['classes']['Monday']['classes'][0]['Subteacher']
    assert not json['data']['classes']['Monday']['classes'][0]['Cancelled']

    # check classes_subteachers
    url = base_url + \
        '/api/schedule_get.json?user=test&key=test&year=2014&week=3'
    page = urllib.urlopen(url).read()
    
    json = sj.loads(page)
    
    assert json['data']['classes']['Monday']['date'] == '2014-01-13'
    assert not json['data']['classes']['Monday']['holiday']
    
    assert json['data']['classes']['Monday']['classes'][0]['TeacherID'] == 3
    assert json['data']['classes']['Monday']['classes'][0]['Teacher'] == \
        'Aimee Garcias'
    assert json['data']['classes']['Monday']['classes'][0]['TeacherID2'] == 2
    assert json['data']['classes']['Monday']['classes'][0]['Teacher2'] == \
        'Pietje Puk'
    assert json['data']['classes']['Monday']['classes'][0]['Subteacher']
    assert not json['data']['classes']['Monday']['classes'][0]['Cancelled']
        
    # check cancelled
    url = base_url + \
        '/api/schedule_get.json?user=test&key=test&year=2014&week=4'
    page = urllib.urlopen(url).read()
    
    json = sj.loads(page)

    assert json['data']['classes']['Monday']['date'] == '2014-01-20'
    assert not json['data']['classes']['Monday']['holiday']
    assert json['data']['classes']['Monday']['classes'][0]['Cancelled']
        
    # check holiday
    url = base_url + \
        '/api/schedule_get.json?user=test&key=test&year=2014&week=5'
    page = urllib.urlopen(url).read()
    
    json = sj.loads(page)

    assert json['data']['classes']['Monday']['date'] == '2014-01-27'
    assert json['data']['classes']['Monday']['holiday']
        
